"""Weighted quantiles."""

from typing import Literal, TypeVar, overload

import numba
import numpy as np

from neo_ls_svm._typing import FloatMatrix, FloatTensor, FloatVector

F = TypeVar("F", np.float32, np.float64)


@numba.jit(nopython=True, nogil=True, parallel=True, fastmath=True, cache=True)
def _parallel_interp(q: FloatVector[F], p: FloatMatrix[F], a: FloatMatrix[F]) -> FloatMatrix[F]:
    out = np.empty((a.shape[0], len(q)), dtype=a.dtype)
    for i in numba.prange(a.shape[0]):
        out[i, :] = np.interp(q, p[i, :], a[i, :])
    return out


@overload
def weighted_quantile(
    a: FloatTensor[F], w: FloatTensor[F], q: float | FloatVector[F], axis: int
) -> FloatTensor[F]:
    ...


@overload
def weighted_quantile(
    a: FloatTensor[F], w: FloatTensor[F], q: float | FloatVector[F], axis: Literal[None]
) -> FloatVector[F]:
    ...


def weighted_quantile(
    a: FloatTensor[F],
    w: FloatTensor[F],
    q: float | FloatVector[F],
    axis: int | None = None,
) -> FloatTensor[F] | FloatVector[F]:
    """Compute the weighted q'th quantile of the data along the specified axis."""
    assert a.ndim == w.ndim, "Array and weights must have the same number of dimensions"
    assert axis is None or (0 <= axis < a.ndim), "Axis must be one of the array's dimensions"
    assert np.all(w >= 0), "Weights must be nonnegative"
    a, w = np.ascontiguousarray(a), np.ascontiguousarray(w)
    w = np.broadcast_to(w, a.shape)  # TODO: Short-circuit to np.quantile if weights are constant.
    q = np.ravel(np.asarray([q])).astype(a.dtype)
    if axis is not None:
        a, w = np.moveaxis(a, axis, -1), np.moveaxis(w, axis, -1)
        a_shape = a.shape
        a, w = np.reshape(a, [-1, a.shape[-1]]), np.reshape(w, [-1, w.shape[-1]])
        idx = np.argsort(a, axis=1)
        a, w = np.take_along_axis(a, idx, axis=1), np.take_along_axis(w, idx, axis=1)
        p = np.cumsum(w, axis=1)
        w_sum = p[:, [-1]].copy()
        p_lower = (p - w) / w_sum
        p_upper = p / w_sum
        tensor: FloatTensor[F] = (
            _parallel_interp(q, p_lower, a) + _parallel_interp(q, p_upper, a)
        ) / 2
        tensor = np.reshape(tensor, a_shape[:-1] + (len(q),))
        tensor = np.moveaxis(tensor, -1, axis)
        output = tensor
    else:
        a, w = np.ravel(a), np.ravel(w)
        idx = np.argsort(a)
        a, w = a[idx], w[idx]
        p = np.cumsum(w)
        # The definition below produces the desired result of 0.5 for the toy example a=(0, 1, 1),
        # w=(2, 1, 1), q=0.5, while the more standard `p = (p - 0.5 * w) / p[-1]` does not.
        p_lower = (p - w) / p[-1]
        p_upper = p / p[-1]
        vector: FloatVector[F] = (
            0.5 * np.interp(q, p_lower, a) + 0.5 * np.interp(q, p_upper, a)
        ).astype(a.dtype)
        output = vector
    return output



"""Neo LS-SVM types."""

from typing import TypeAlias, TypeVar

import numpy as np
import numpy.typing as npt

C = TypeVar("C", np.complex64, np.complex128)
F = TypeVar("F", np.float32, np.float64)
K = TypeVar("K", np.int8, np.uint8, np.int16, np.uint16, np.int32, np.uint32, np.int64, np.intp)
N = TypeVar("N", np.float32, np.float64, np.int32, np.int64, np.intp)

# TODO: https://github.com/numpy/numpy/issues/16544
ComplexVector: TypeAlias = npt.NDArray[C]
ComplexMatrix: TypeAlias = npt.NDArray[C]
ComplexTensor: TypeAlias = npt.NDArray[C]

FloatVector: TypeAlias = npt.NDArray[F]
FloatMatrix: TypeAlias = npt.NDArray[F]
FloatTensor: TypeAlias = npt.NDArray[F]

GenericVector: TypeAlias = npt.NDArray[np.generic]
GenericMatrix: TypeAlias = npt.NDArray[np.generic]
GenericTensor: TypeAlias = npt.NDArray[np.generic]

IntegerVector: TypeAlias = npt.NDArray[K]
IntegerMatrix: TypeAlias = npt.NDArray[K]
IntegerTensor: TypeAlias = npt.NDArray[K]

NumberVector: TypeAlias = npt.NDArray[N]
NumberMatrix: TypeAlias = npt.NDArray[N]
NumberTensor: TypeAlias = npt.NDArray[N]



"""Affine feature map."""

from functools import cached_property
from typing import Any, TypeVar, cast

import numpy as np
import numpy.typing as npt
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_array
from sklearn.utils.validation import _check_feature_names_in

from neo_ls_svm._typing import FloatMatrix, FloatVector

F = TypeVar("F", np.float32, np.float64)


class AffineFeatureMap(BaseEstimator, TransformerMixin):
    """Affine feature map.

    Applies the transformation (x - shift) @ diag(1 / scale) @ A to the input row x for a given
    vector shift, vector scale, and matrix A.

    By setting append_features=True the transformed features are appended to the given features if
    the transformation matrix A is not None.
    """

    def __init__(
        self,
        *,
        scale: FloatVector[F],
        shift: FloatVector[F],
        A: FloatMatrix[F] | None = None,
        append_features: bool = False,
    ):
        self.scale = scale  # type: ignore[var-annotated]
        self.shift = shift  # type: ignore[var-annotated]
        self.A = A  # type: ignore[var-annotated]
        self.append_features = append_features

    def fit(
        self,
        X: FloatMatrix[F],
        y: FloatVector[F] | None = None,
        sample_weight: FloatVector[F] | None = None,
    ) -> "AffineFeatureMap":
        """Fit this transformer."""
        X = check_array(X)
        self.n_features_in_ = X.shape[1]
        scale = np.reshape(getattr(self, "scale_", self.scale), (-1, X.shape[1]))
        shift = np.reshape(getattr(self, "shift_", self.shift), (-1, X.shape[1]))
        A = getattr(self, "A_", self.A)
        assert scale.dtype == shift.dtype, "The scale and shift must have the same dtype"
        assert not np.any(scale == 0), "The scale may not be zero"
        assert np.all(np.isfinite(scale)), "The scale must be finite"
        assert np.all(np.isfinite(shift)), "The shift must be finite"
        assert (
            X.shape[1] == scale.shape[1]
        ), "The scale must be compatible with the number of features"
        assert (
            X.shape[1] == shift.shape[1]
        ), "The shift must be compatible with the number of features"
        if A is not None:
            assert (
                A.dtype == scale.dtype
            ), "The matrix A must have the same dtype as the scale and shift"
            assert (
                X.shape[1] == A.shape[0]
            ), "The matrix A must have rows equal to the number of features in X"
            assert np.all(np.isfinite(A)), "The matrix A must be finite"
        return self

    def transform(self, X: FloatMatrix[F]) -> FloatMatrix[F]:
        """Transform the given data with this transformer."""
        # Check the input.
        X = check_array(X)
        # Get the fitted version of the parameters, if they exist.
        scale = np.reshape(getattr(self, "scale_", self.scale), (-1, X.shape[1]))
        shift = np.reshape(getattr(self, "shift_", self.shift), (-1, X.shape[1]))
        A = getattr(self, "A_", self.A)
        # Transform the input.
        X_transformed: FloatMatrix[F] = (
            (X - shift) / scale
            if A is None
            else (
                X @ (A / scale.T) - shift @ (A / scale.T)  # Memory optimisation.
                if A.shape[1] < A.shape[0]
                else (X - shift) @ (A / scale.T)
            )
        ).astype(X.dtype)
        if self.append_features and A is not None:
            X_transformed = np.hstack((X, X_transformed))
        return X_transformed

    @cached_property
    def pseudo_inverse(self) -> FloatMatrix[F] | None:
        """The pseudo-inverse of this transform's matrix A."""
        return np.linalg.pinv(self.A) if self.A is not None else None

    def inverse_transform(self, X_transformed: FloatMatrix[F]) -> FloatMatrix[F]:
        """Approximately invert this transformation."""
        X: FloatMatrix[F] = check_array(X_transformed)
        scale = np.reshape(getattr(self, "scale_", self.scale), (-1, X.shape[1]))
        shift = np.reshape(getattr(self, "shift_", self.shift), (-1, X.shape[1]))
        A = getattr(self, "A_", self.A)
        if self.append_features and A is not None:
            X = X[:, : A.shape[0]]
        else:
            if A is not None:
                pinvA = cast(FloatMatrix[F], self.pseudo_inverse)
                X = X @ pinvA
            X = (X * scale + shift).astype(X.dtype)
        return X

    def get_feature_names_out(
        self, input_features: npt.ArrayLike | None = None
    ) -> npt.NDArray[np.object_]:
        """Get output feature names for transformation."""
        A = getattr(self, "A_", self.A)
        input_features_array = cast(
            npt.NDArray[np.object_], _check_feature_names_in(self, input_features)
        )
        output_features: npt.NDArray[np.object_] = (
            input_features_array + "_shifted_scaled"
            if A is None
            else np.array(
                [f"{','.join(list(input_features_array))}_affine_map"] * A.shape[1],
                dtype=object,
            )
        )
        if self.append_features and A is not None:
            output_features = np.hstack((input_features_array, output_features))
        return output_features

    def _more_tags(self) -> dict[str, Any]:
        # https://scikit-learn.org/stable/developers/develop.html#estimator-tags
        return {"preserves_dtype": [np.float64, np.float32]}



"""Random feature maps.

This module implements feature maps φ: Rᵈ -> Cᴰ so that φ(x)'φ(y) approximates the Gaussian kernel
k(x, y) := exp(- ||x - y||² / (2σ²)), or more generally k(x, y) := exp(- ||A(x - y)||² / 2) for a
matrix A.

These approximating feature maps also include transformers to automatically learn an affine
transformation A that optimally separates the examples.

Notable kernel-approximating feature maps are Random Fourier Features [1][2][3], Orthogonal Random
Features [4], and Spherical Structured Features [5]. For a full overview, see the survey [6].

[1] https://people.eecs.berkeley.edu/~brecht/papers/07.rah.rec.nips.pdf
[2] https://arxiv.org/abs/1408.3060
[3] https://gregorygundersen.com/blog/2019/12/23/random-fourier-features
[4] https://arxiv.org/abs/1610.09072
[5] http://proceedings.mlr.press/v70/lyu17a/lyu17a.pdf
[6] https://arxiv.org/abs/2004.11154
"""

from abc import ABC, abstractmethod
from functools import cached_property
from typing import TypeVar

import numba
import numpy as np
import numpy.typing as npt
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_random_state

from neo_ls_svm._affine_feature_map import AffineFeatureMap
from neo_ls_svm._affine_separator import AffineSeparator
from neo_ls_svm._typing import ComplexMatrix, FloatMatrix, FloatVector

C = TypeVar("C", np.complex64, np.complex128)
F = TypeVar("F", np.float32, np.float64)


@numba.jit(nopython=True, nogil=True, parallel=True, fastmath=True, cache=True)
def _ztz_prod_sinc_zmz(Z: FloatMatrix[F], *, fast_approx: bool = False) -> FloatMatrix[F]:
    """Compute 1/d Z'Z * [Πₖ sinc(Zₖᵢ - Zₖⱼ)]ᵢⱼ)."""
    d, D = Z.shape
    if fast_approx:
        # Slower approx: np.diag(np.sum(Z * Z, axis=0)) / d = np.diag(np.diag(Z.T @ Z)) / d.
        return np.eye(D, dtype=Z.dtype)
    C = Z.T @ Z
    eps = np.finfo(Z.dtype).eps
    for k in range(d):
        for i in numba.prange(D):
            for j in numba.prange(i):
                dz = Z[k, i] - Z[k, j]
                if np.abs(dz) > eps:
                    C[i, j] *= np.sin(dz) / dz
    C_scaled: FloatMatrix[F] = (np.tril(C) + np.tril(C, -1).T) / d
    return C_scaled


class KernelApproximatingFeatureMap(ABC, BaseEstimator, TransformerMixin):
    """Abstract kernel-approximating feature map."""

    def __init__(
        self,
        affine_feature_map: AffineFeatureMap | None = None,
        num_features: int = 512,
        random_state: int | np.random.RandomState | None = 42,
    ):
        self.num_features, self.D = num_features, num_features
        self.affine_feature_map = affine_feature_map or AffineSeparator()
        self.random_state = random_state

    @cached_property
    @abstractmethod
    def complexity_matrix(self) -> FloatMatrix[F]:
        """Compute the complexity regularisation matrix.

        This regularisation matrix is a complement to the classic maximal margin (classification) or
        ridge regression (regression) identity matrix. The complexity matrix directly penalises the
        complexity of the prediction surface φ(x)'w + b by integrating the norm of its normal vector
        ∇ₓφ(x)'w over the normalised feature space (-1, 1)ᵈ:

            φ(x) := exp(1j Z'x) / √D
            ∇ₓφ(x) = diag(exp(1j Z'x)) Z' / √D
            ∇ₓφ(x)∇ₓφ(x)' = diag(exp(1j Z'x)) Z'Z diag(exp(-1j Z'x)) / D

            ∫||∇ₓφ(x)'w||²dx
            = ∫ w'∇ₓφ(x)∇ₓφ(x)'w dx
            = w' ∫∇ₓφ(x)∇ₓφ(x)'dx w
            = w' ∫ Z'Z/D * [exp(1j Z₌ᵢ'x - 1j Z₌ⱼ'x)]ᵢⱼ dx  where  * is the Hadamard product
            = w' (Z'Z/D * ∫ [exp(1j (Z₌ᵢ - Z₌ⱼ)'x)]ᵢⱼ dx) w  where  Z₌ᵢ is the i'th column of Z
            = w' (Z'Z * [2ᵈ/D Πₖ sinc(Zₖᵢ - Zₖⱼ)]ᵢⱼ) w
            ~ w' (Z'Z * [Πₖ sinc(Zₖᵢ - Zₖⱼ)]ᵢⱼ) w

        We also extend the complexity matrix with an entry on the diagonal to also shrink the bias
        term b, and normalise the matrix so that it is invariant to the number of input features d
        and the number of output features D.
        """
        ...

    @abstractmethod
    def fit(
        self,
        X: FloatMatrix[F],
        y: FloatVector[F] | None = None,
        sample_weight: FloatVector[F] | None = None,
    ) -> "KernelApproximatingFeatureMap":
        """Fit this transformer."""
        self.affine_feature_map.fit(X, y, sample_weight)
        self.n_features_in_ = X.shape[1]
        return self

    @abstractmethod
    def transform(self, X: FloatMatrix[F]) -> ComplexMatrix[C]:
        """Transform the given data with this transformer."""
        ...


class RandomFourierFeatures(KernelApproximatingFeatureMap):
    """Random Fourier Features."""

    @classmethod
    def _fourier_features(
        cls, d: int, D: int, dtype: npt.DTypeLike, random_state: int | np.random.RandomState | None
    ) -> FloatMatrix[F]:
        # Draw elements of Z ∈ Rᵈˣᴰ from N(0, 1).
        generator = check_random_state(random_state)
        Z: FloatMatrix[F] = generator.randn(d, D).astype(dtype)
        return Z

    @cached_property
    def complexity_matrix(self) -> FloatMatrix[F]:
        """Compute this feature map's complexity matrix."""
        # Compute the complexity matrix using a fast diagonal approximation.
        C: FloatMatrix[F] = np.eye(self.D + 1, dtype=self.Z_.dtype)
        C[:-1, :-1] = _ztz_prod_sinc_zmz(self.Z_, fast_approx=True)
        return C

    def fit(
        self,
        X: FloatMatrix[F],
        y: FloatVector[F] | None = None,
        sample_weight: FloatVector[F] | None = None,
    ) -> "RandomFourierFeatures":
        """Fit this transformer."""
        # Fit an affine transform.
        super().fit(X, y, sample_weight)
        # Create a linear transformation Z ∈ Rᵈˣᴰ and update the affine transform.
        A = getattr(self.affine_feature_map, "A_", self.affine_feature_map.A)
        d = A.shape[1] if A is not None else X.shape[1]
        self.Z_: FloatMatrix[F] = self._fourier_features(d, self.D, X.dtype, self.random_state)
        self.affine_feature_map.A_ = A @ self.Z_ if A is not None else self.Z_
        return self

    def transform(self, X: FloatMatrix[F]) -> ComplexMatrix[C]:
        """Transform a feature matrix X ∈ Rⁿˣᵈ into φ(X) ∈ Cⁿˣᴰ⁺¹ so that φ(X)ᵢ := [φ(xᵢ)' 1].

        Notice that we can choose to solve an LS-SVM in the primal or dual space using the
        push-through identity (γ𝕀 + AB)⁻¹ A = A (γ𝕀 + BA)⁻¹:

            argmin ||φ(X)β̂ - y||² + γ||β̂||²
            = (γ𝕀 + φ(X)'φ(X))⁻¹ φ(X)'y
            = φ(X)' (γ𝕀 + φ(X)φ(X)')⁻¹y  with the identity  (γ𝕀 + AB)⁻¹ A = A (γ𝕀 + BA)⁻¹
            = φ(X)'α̂  where  α̂ := (γ𝕀 + φ(X)φ(X)')⁻¹y = (γ𝕀 + k(xᵢ, xⱼ))⁻¹y

        This means that k(x, y) = φ(x)'φ(y) by definition. Now we look for a φ(x) so that k(x, y) =
        φ(x)'φ(y) for the Gaussian kernel k(x, y) = exp(- ||y - x||² / 2). If we take h(x) :=
        exp(1j ω'Re{x}) for a standard normally distributed variable ω, then we see that [1]:

            = Eω[h(x)'h(y)]
            = Eω[exp(-1j Re{x}'ω) exp(1j ω'Re{y})]
            = ∫ p(ω) exp(-1j/2 (ω'x + x'ω)) exp(1j/2 (ω'y + y'ω)) dω
            = ∫ p(ω) exp(1j/2 (ω'y + y'ω - x'ω - ω'x)) dω
            = ∫ p(ω) exp(-1/2 [- 1j ω'(y - x) - 1j (y - x)'ω]) dω
            = (2π)^(-D/2) ∫ exp(- ||ω||²/2) exp(-1/2 [- 1j ω'(y - x) - 1j(y - x)'ω]) dω
            = (2π)^(-D/2) ∫ exp(-1/2 [ω'ω - 1j ω'(y - x) - 1j(y - x)'ω]) dω
            = (2π)^(-D/2)
              ∫ exp(-1/2 [ω'ω - 1j ω'(y - x) - 1j(y - x)'ω - ||y - x||²] - 1/2 ||y - x||²) dω
            = (2π)^(-D/2) exp(- ||y - x||² / 2) ∫ exp(-1/2 (ω - 1j(y - x))'(ω - 1j(y - x))) dω
            = exp(- ||y - x||² / 2)

        Notes
        -----
        1. The resulting features are complex-valued. There does exist a real version
           [cos(XZ) sin(XZ)] / √D of this feature map. However, this doubles the size of the linear
           system, which results in an increase of 2³ = 8 work compared to the factor of ~4 for
           working with a complex-valued system of half the size.
        2. The rows of the output φ(X) are of the form [φ(xᵢ)' 1] where x is a row of X transposed
           into a column vector and φ(x) := exp(1j Z'x) / √D so that the model f(x) := φ(x)'w + b
           can be computed for multiple examples X as φ(X)[w; b].

        References
        ----------
        [1] https://gregorygundersen.com/blog/2019/12/23/random-fourier-features/#a1-gaussian-kernel-derivation
        """
        # Apply an affine transformation to X.
        X = self.affine_feature_map.transform(X)
        # Apply the complex feature map to X.
        φ: ComplexMatrix[C] = np.empty(
            (X.shape[0], self.D + 1),
            dtype=np.complex64 if X.dtype == np.float32 else np.complex128,
        )
        φ[:, :-1] = np.exp(-1j * X, dtype=φ.dtype) / np.sqrt(self.D)
        φ[:, -1] = 1
        return φ


class OrthogonalRandomFourierFeatures(RandomFourierFeatures):
    """Orthogonal Random Fourier Features."""

    @classmethod
    def _fourier_features(
        cls, d: int, D: int, dtype: npt.DTypeLike, random_state: int | np.random.RandomState | None
    ) -> FloatMatrix[F]:
        # Draw elements of Z ∈ Rᵈˣᴰ from N(0, 1).
        generator = check_random_state(random_state)
        Z: FloatMatrix[F] = generator.randn(d, D).astype(dtype)
        # Then orthonormalise its columns.
        for j in range(0, D, d):
            Q, _ = np.linalg.qr(Z[:, j : j + d])
            Z[:, j : j + d] = Q
        # And compensate for the loss of magnitude relative to standard normal vectors.
        S = np.sqrt(generator.chisquare(d, size=(1, Z.shape[1])).astype(dtype))
        Z *= S
        return Z



"""Quantisation of numerical features."""

from typing import Any, Literal, TypeVar, cast, overload

import numba
import numpy as np
import numpy.typing as npt
from sklearn.base import BaseEstimator, TransformerMixin, check_array
from sklearn.utils.validation import _check_feature_names_in

from neo_ls_svm._typing import FloatVector, GenericVector, IntegerVector, NumberMatrix, NumberVector

F = TypeVar("F", np.float32, np.float64)
M = TypeVar("M", np.float32, np.float64, np.int32, np.int64, np.intp)
N = TypeVar("N", np.float32, np.float64, np.int32, np.int64, np.intp)


@numba.jit(nopython=True, nogil=True, parallel=False, fastmath=True, cache=True)
def _next_knot(
    x: FloatVector[F], y: IntegerVector[np.int64], knot: int, max_bin_error: int, max_bin_size: int
) -> tuple[int, int]:
    # Find next_knot which is the first index after knot for which the linear interpolation between
    # [x[knot], x[next_knot]) has an error larger than max_bin_error, or the number of samples in
    # [x[knot], x[next_knot]) is larger than max_bin_size.
    min_a, max_a = 0.0, np.inf
    for next_knot in range(knot + 1, len(x)):
        # The bin count includes all data points in [knot, next_knot).
        bin_count = int(y[next_knot - 1] - y[knot - 1] if knot > 0 else y[next_knot - 1])
        # Return next_knot if we have accumulated more data points than than the maximum bin size.
        if bin_count > max_bin_size:
            break
        # Continue if the current bin consists of only one knot.
        if next_knot == knot + 1:
            continue
        # Update the minimum and maximum tangents that would cause the error within the current bin
        # to exceed the max error.
        dx, dy = x[next_knot - 1] - x[knot], y[next_knot - 1] - y[knot]
        max_a = min(max_a, (dy + max_bin_error) / dx)
        min_a = max(min_a, (dy - max_bin_error) / dx)
        # If the tangent exceeds the minimum or maximum error, return the bin.
        a = dy / dx
        if not (min_a <= a <= max_a):
            break
    return next_knot, bin_count


@numba.jit(nopython=True, nogil=True, parallel=False, fastmath=True, cache=True)
def _prev_knot(
    x: FloatVector[F], y: IntegerVector[np.int64], knot: int, max_bin_error: int, max_bin_size: int
) -> tuple[int, int]:
    # Find prev_knot which is the first index before knot for which the linear interpolation between
    # [x[prev_knot], x[knot]) has an error larger than max_bin_error, or the number of samples in
    # [x[prev_knot], x[knot]) is larger than max_bin_size.
    min_a, max_a = 0.0, np.inf
    for prev_knot in range(knot - 1, -1, -1):
        # The bin count includes all data points in [prev_knot, knot).
        bin_count = int(y[knot - 1] - y[prev_knot - 1] if prev_knot > 0 else y[knot - 1])
        # Return next_knot if we have accumulated more data points than than the maximum bin size.
        if bin_count > max_bin_size:
            break
        # Continue if the current bin consists of only one knot.
        if knot == prev_knot + 1:
            continue
        # Update the minimum and maximum tangents that would cause the error within the current bin
        # to exceed the max error.
        dx, dy = x[knot - 1] - x[prev_knot], y[knot - 1] - y[prev_knot]
        max_a = min(max_a, (dy + max_bin_error) / dx)
        min_a = max(min_a, (dy - max_bin_error) / dx)
        # If the tangent exceeds the minimum or maximum error, return the bin.
        a = dy / dx
        if not (min_a <= a <= max_a):
            break
    return prev_knot, bin_count


@overload
def hist_quantized_ecdf(
    x: NumberVector[N],
    *,
    density: Literal[True],
    max_bin_error: float = ...,
    max_bin_size: float = ...,
) -> tuple[FloatVector[F], FloatVector[F]]:
    ...


@overload
def hist_quantized_ecdf(
    x: NumberVector[N],
    *,
    density: Literal[False],
    max_bin_error: float = ...,
    max_bin_size: float = ...,
) -> tuple[IntegerVector[np.int64], FloatVector[F]]:
    ...


def hist_quantized_ecdf(
    x: NumberVector[N],
    *,
    density: bool = False,
    max_bin_error: float = 0.0125,
    max_bin_size: float = 0.125,
    merge_bin_size: float = 0.025,
) -> tuple[FloatVector[F] | IntegerVector[np.int64], FloatVector[F]]:
    """Compute a vector's histogram by quantizing its empirical cumulative distribution function."""
    # Convert relative error and bin size to absolute error and bin size.
    max_bin_error = int(max_bin_error * len(x))
    max_bin_size = int(max_bin_size * len(x))
    merge_bin_size = int(merge_bin_size * len(x))
    # Compute the empirical cumulative distribution function of x.
    x, counts = np.unique(x, return_counts=True)  # Sorted unique values of x, with counts.
    # Improve output quality by starting with the largest x values to avoid ending up with a tiny
    # residual bin with only a few samples. This conditioning could be improved upon by making the
    # algorithm double-sided.
    y = np.cumsum(counts)
    # Quantize x, y optimally one bin at a time.
    x_, y_ = np.append(x, np.inf), np.append(y, np.iinfo(y.dtype).max)
    x_, y_ = np.insert(x_, 0, -np.inf), np.insert(y_, 0, 0)
    knot_left = 1
    knot_right = len(x_) - 1
    bin_edges_left = [x[0]]
    bin_edges_right = [x[-1]]
    hist_left: list[int] = []
    hist_right: list[int] = []
    while knot_left < knot_right:
        # Determine next not from the left, and previous knot from the right.
        knot_left_prev = knot_left
        knot_right_prev = knot_right
        knot_left, bin_count_left = _next_knot(x_, y_, knot_left, max_bin_error, max_bin_size)
        knot_right, bin_count_right = _prev_knot(x_, y_, knot_right, max_bin_error, max_bin_size)
        # Add bin counts.
        hist_left.append(bin_count_left)
        hist_right.insert(0, bin_count_right)
        # Add bin edges.
        bin_edges_left.append(
            (x_[knot_left] + x_[knot_left - 1]) / 2 if knot_left > 0 else x_[knot_left]
        )
        bin_edges_right.insert(
            0, (x_[knot_right] + x_[knot_right - 1]) / 2 if knot_right > 0 else x_[knot_right]
        )
        # Check if we should stop.
        if knot_left == knot_right:
            bin_edges = bin_edges_left + bin_edges_right[1:]
            hist = hist_left + hist_right
            break
        if knot_left > knot_right:
            hist = (
                hist_left[:-1]
                + [y[-1] - np.sum(hist_left[:-1]) - np.sum(hist_right[1:])]
                + hist_right[1:]
            )
            bin_edges = bin_edges_left[:-1] + bin_edges_right[1:]
            break
        if y_[knot_right - 1] - y_[knot_left - 1] <= merge_bin_size:
            knot_center_left = int(np.floor((knot_left + knot_right) / 2))
            knot_center_right = int(np.ceil((knot_left + knot_right) / 2))
            bin_edge_center = (x_[knot_center_left] + x_[knot_center_right]) / 2
            hist = (  # TODO: Cover this with tests. E.g. assert np.sum(hist) == y[-1].
                hist_left[:-1]
                + [y_[knot_center_left] - y_[knot_left_prev - 1]]
                + [y_[knot_right_prev - 1] - y_[knot_center_right - 1]]
                + hist_right[1:]
            )
            bin_edges = bin_edges_left[:-1] + [bin_edge_center] + bin_edges_right[1:]
            break
    # Convert counts to a density if requested.
    floating_dtype: npt.DTypeLike = x.dtype if np.issubdtype(x.dtype, np.floating) else np.float64
    hist_arr = (np.array(hist) / y[-1]).astype(floating_dtype) if density else np.array(hist)
    bin_edges_arr = np.array(bin_edges).astype(floating_dtype)
    return hist_arr, bin_edges_arr


class Quantizer(BaseEstimator, TransformerMixin):
    """Quantizing encoder for numerical features.

    Maps numerical features to [1, num_bins] by quantizing them into dynamically sized bins.
    """

    def __init__(
        self,
        *,
        max_bin_error: float = 0.0125,
        max_bin_size: float = 0.125,
        append_invfreq: bool = False,
        dtype: npt.DTypeLike = np.intp,
    ):
        self.max_bin_error = max_bin_error
        self.max_bin_size = max_bin_size
        self.append_invfreq = append_invfreq
        self.dtype = dtype
        if append_invfreq and not np.issubdtype(dtype, np.floating):
            self.dtype = np.float32

    def fit(self, X: NumberMatrix[N], y: Any = None) -> "Quantizer":
        """Fit this transformer."""
        X = check_array(X)
        self.n_features_in_ = X.shape[1]
        self.X_hist_: list[IntegerVector[np.int64]] = []
        self.X_bin_edges_ = []
        for j in range(X.shape[1]):
            # Compute a variable-width histogram by quantizing X[:, j]'s ECDF.
            Xj_hist, Xj_bin_edges = hist_quantized_ecdf(
                X[:, j],
                density=False,
                max_bin_error=self.max_bin_error,
                max_bin_size=self.max_bin_size,
            )
            self.X_hist_.append(Xj_hist)
            self.X_bin_edges_.append(Xj_bin_edges)
        return self

    def transform(self, X: NumberMatrix[N]) -> NumberMatrix[M]:
        """Transform the given data with this transformer."""
        X_transformed = np.empty(
            (X.shape[0], (1 + self.append_invfreq) * X.shape[1]), dtype=self.dtype
        )
        for j in range(X.shape[1]):
            Xj_bin_indices = np.clip(
                np.searchsorted(self.X_bin_edges_[j], X[:, j], side="right") - 1,
                0,
                len(self.X_bin_edges_[j]) - 2,
            )
            X_transformed[:, j] = Xj_bin_indices
            if self.append_invfreq:
                X_transformed[:, X.shape[1] + j] = (
                    1 / len(self.X_hist_[j]) / self.X_hist_[j][Xj_bin_indices]
                )
        return X_transformed

    def get_feature_names_out(
        self, input_features: npt.ArrayLike | None = None
    ) -> npt.NDArray[np.object_]:
        """Get output feature names for transformation."""
        input_features_array = cast(
            npt.NDArray[np.object_], _check_feature_names_in(self, input_features)
        )
        output_features: npt.NDArray[np.object_] = np.array(input_features_array) + "_quantized"
        if self.append_invfreq:
            output_features = np.hstack(
                (output_features, np.array(input_features_array) + "_invfreq")
            )
        return output_features


def sample_bins_quantized_ecdf(x: GenericVector, **kwargs: Any) -> IntegerVector[np.intp]:
    """Compute optimal sample bins of a vector by quantizing its ECDF."""
    x_unique, x = np.unique(x, return_inverse=True)
    if len(x_unique) <= np.ceil(np.sqrt(len(x))):
        return x
    quantizer = Quantizer(dtype=np.intp, **kwargs)
    sample_bins: IntegerVector[np.intp] = quantizer.fit_transform(x[:, np.newaxis]).ravel()
    return sample_bins


def sample_weights_quantized_ecdf(x: GenericVector, **kwargs: Any) -> FloatVector[F]:
    """Compute optimal sample weights of a vector by quantizing its ECDF."""
    dtype: npt.DTypeLike = x.dtype if np.issubdtype(x.dtype, np.floating) else np.float64
    x_unique, x, x_counts = np.unique(x, return_inverse=True, return_counts=True)
    if len(x_unique) <= np.ceil(np.sqrt(len(x))):
        return x_counts[x] / np.sum(x_counts)
    quantizer = Quantizer(append_invfreq=True, dtype=dtype, **kwargs)
    sample_weights: FloatVector[F] = quantizer.fit_transform(x[:, np.newaxis])[:, 1]
    return sample_weights



"""Supervised affine normalizer."""

from typing import TypeVar, cast

import numpy as np
from sklearn.utils import check_consistent_length, check_X_y

from neo_ls_svm._affine_feature_map import AffineFeatureMap
from neo_ls_svm._quantizer import sample_bins_quantized_ecdf
from neo_ls_svm._typing import FloatMatrix, FloatVector
from neo_ls_svm._weighted_quantile import weighted_quantile

F = TypeVar("F", np.float32, np.float64)


class AffineNormalizer(AffineFeatureMap):
    """Supervised affine normalizer.

    Applies the transformation (x - shift) @ diag(1 / scale) to the input row x so that:

        1. shift centers the features x so that the class bins are optimally separated, and
        2. scale scales the shifted features x so that the difference between two samples from
           different class bins is the separability between those class bins.

    To compute the shift and scale, the samples are first quantized into class bins. Then, each
    class bin's median and standard deviation are computed.

    A feature's shift is determined as a weighted sum of shifts, where each shift is the optimal
    shift for a pair of class bins. The weights of these shifts are the geometric mean of the class
    bins' sample weight and the separability of that pair of class bins.

    The separability of a pair of class bins is defined as:

        np.abs(X_bins_μ[j] - X_bins_μ[i]) / (X_bins_σ[i] + X_bins_σ[j])

    where X_bins_μ[k] and X_bins_σ[k] are the k'th class bin's center and standard deviation,
    respectively.

    The scale is analogously determined as a weighted sum of scales, where each scale is the sum of
    a pair of class bin standard deviations. In this way, the difference between two features after
    transformation corresponds to the separability between their class bins.
    """

    def __init__(self, *, append_features: bool = False) -> None:
        self.shift = 0.0
        self.scale = 1.0
        self.A = None
        self.append_features = append_features

    def fit(
        self,
        X: FloatMatrix[F],
        y: FloatVector[F] | None = None,
        sample_weight: FloatVector[F] | None = None,
    ) -> "AffineFeatureMap":
        """Fit this transformer."""
        # Validate X and y.
        X, y = check_X_y(X, y, dtype=(np.float64, np.float32))
        y = np.ravel(np.asarray(y)).astype(X.dtype)
        # Use uniform sample weights if none are provided.
        sample_weight_ = cast(
            FloatVector[F],
            np.ones(y.shape) if sample_weight is None else np.ravel(np.asarray(sample_weight)),
        ).astype(y.dtype)
        check_consistent_length(y, sample_weight_)
        # Separate X into bins according to y.
        y_quantized = sample_bins_quantized_ecdf(y)
        bins = [y_quantized == i for i in range(np.min(y_quantized), np.max(y_quantized) + 1)]
        X_bins = [X[bin, :] for bin in bins]  # noqa: A001
        n_bins = [np.sum(sample_weight_[bin]) for bin in bins]  # noqa: A001
        s_bins = [sample_weight_[np.newaxis, bin] / np.sum(sample_weight_[bin]) for bin in bins]  # noqa: A001
        # Exit early if there is only one bin.
        self.shift_: FloatVector[F]
        self.scale_: FloatVector[F]
        if len(X_bins) <= 1:
            self.shift_ = np.zeros((1, X.shape[1]), dtype=X.dtype)
            self.scale_ = np.ones((1, X.shape[1]), dtype=X.dtype)
            super().fit(X, y, sample_weight_)
            return self
        # Compute an optimal shift and scale.
        X_bins_μ: list[FloatVector[F]] = [
            weighted_quantile(X_bin, s_bin.T, 0.5, axis=0)
            for X_bin, s_bin in zip(X_bins, s_bins, strict=False)
        ]
        X_bins_σ: list[FloatVector[F]] = [
            s_bin @ np.abs(X_bin - X_bin_μ)
            for X_bin, s_bin, X_bin_μ in zip(X_bins, s_bins, X_bins_μ, strict=False)
        ]
        sign = np.zeros((1, X.shape[1]), dtype=X.dtype)
        sum_w: FloatVector[F] = np.zeros((1, X.shape[1]), dtype=X.dtype)
        self.shift_ = np.zeros((1, X.shape[1]), dtype=X.dtype)
        self.scale_ = np.zeros((1, X.shape[1]), dtype=X.dtype)
        for i in range(len(X_bins_μ) - 1):
            for j in range(i + 1, len(X_bins_μ)):
                # Compute the distance between class bin centroids and the total variance in the two
                # class bins.
                diff_μ: FloatVector[F] = X_bins_μ[j] - X_bins_μ[i]
                sum_σ: FloatVector[F] = np.maximum(X_bins_σ[i] + X_bins_σ[j], np.finfo(X.dtype).eps)
                # Determine the weight of this pair of class bins as the geometric mean of their
                # sample weight and separability.
                separability = np.abs(diff_μ) / sum_σ
                w: float = np.sqrt(
                    (n_bins[i] + n_bins[j]) * (0.5 + separability)
                )  # Regularised GM.
                # For this pair of class bins, compute the optimal separation threshold.
                alpha = np.clip(X_bins_σ[i] / sum_σ, 1e-6, 1.0 - 1e-6)
                self.shift_ = self.shift_ + w * (X_bins_μ[i] + alpha * diff_μ)
                self.scale_ = self.scale_ + w * sum_σ
                sign += w * np.sign(diff_μ)
                sum_w += w
        sign /= sum_w
        self.shift_ = self.shift_ / sum_w  # type: ignore[assignment]
        self.scale_ = self.scale_ / sum_w  # type: ignore[assignment]
        self.scale_[np.sign(sign) < 0] = -self.scale_[np.sign(sign) < 0]
        # Validate the learned parameters.
        super().fit(X, y, sample_weight_)
        return self



"""Neo LS-SVM."""

from typing import Any, Literal, TypeVar, cast

import numpy as np
import numpy.typing as npt
from scipy.linalg import cho_factor, cho_solve, eigh, lu_factor, lu_solve
from sklearn.base import BaseEstimator, clone
from sklearn.isotonic import IsotonicRegression
from sklearn.linear_model import QuantileRegressor
from sklearn.metrics import accuracy_score, r2_score
from sklearn.metrics.pairwise import euclidean_distances, rbf_kernel
from sklearn.utils.validation import check_consistent_length, check_X_y

from neo_ls_svm._affine_feature_map import AffineFeatureMap
from neo_ls_svm._affine_separator import AffineSeparator
from neo_ls_svm._feature_maps import (
    KernelApproximatingFeatureMap,
    OrthogonalRandomFourierFeatures,
)
from neo_ls_svm._typing import (
    ComplexMatrix,
    ComplexVector,
    FloatMatrix,
    FloatTensor,
    FloatVector,
    GenericVector,
)

C = TypeVar("C", np.complex64, np.complex128)
F = TypeVar("F", np.float32, np.float64)


class NeoLSSVM(BaseEstimator):
    """Neo LS-SVM.

    A neo Least-Squares Support Vector Machine with:

        1. ⚡ Linear complexity in the number of training examples with Orthogonal Random Features.
        2. 🚀 Hyperparameter free: zero-cost optimization of the regularisation parameter γ and
             kernel parameter σ.
        3. 🏔️ Adds a new tertiary objective that minimizes the complexity of the prediction surface.
        4. 🎁 Returns the leave-one-out residuals and error for free after fitting.
        5. 🌀 Learns an affine transformation of the feature matrix to optimally separate the
             target's bins.
        6. 🪞 Can solve the LS-SVM both in the primal and dual space.
        7. 🌡️ Isotonically calibrated `predict_proba` based on the leave-one-out predictions.
        8. 🎲 Asymmetric conformal Bayesian confidence intervals for classification and regression.
    """

    def __init__(  # noqa: PLR0913
        self,
        *,
        primal_feature_map: KernelApproximatingFeatureMap | Literal["auto"] = "auto",
        dual_feature_map: AffineSeparator | Literal["auto"] = "auto",
        dual: bool | Literal["auto"] = "auto",
        estimator_type: Literal["auto", "classifier", "regressor"] = "auto",
        random_state: int | np.random.RandomState | None = 42,
    ) -> None:
        self.primal_feature_map = primal_feature_map
        self.dual_feature_map = dual_feature_map
        self.dual = dual
        self.random_state = random_state
        self.estimator_type = estimator_type

    def _optimize_β̂_γ(
        self,
        φ: ComplexMatrix[C],
        y: FloatVector[F],
        s: FloatVector[F],
        C: FloatMatrix[F],
    ) -> tuple[ComplexVector[C], float]:
        """Find β̂ = argmin ||S(φ(X)β̂ - y)||² + γβ̂'Cβ̂ and γ that minimises the leave-one-out error.

        First, we solve min ||S(φ(X)β̂ - y)||² + γβ̂'Cβ̂ for β̂ as a function of γ::

            (γC + φ(X)'SSφ(X)) β̂ = φ(X)'S Sy
            (γ𝕀 + C⁻¹φ(X)'SSφ(X)) β̂ = C⁻¹φ(X)'S Sy
            (γQQ⁻¹ + QΛQ⁻¹) β̂ = C⁻¹φ(X)'S Sy  where  φ(X)'SSφ(X)Q = CQΛ
            β̂ = Q(γ𝕀 + Λ)⁻¹Q⁻¹ C⁻¹φ(X)'S Sy

        The entries of β̂ are rational polynomials of γ: Q diag([rᵢ(γ)]ᵢ) Q⁻¹ C⁻¹φ(X)'SSy. The
        unweighted leave-one-out residuals e⁽ˡᵒᵒ⁾ can be derived by analogy to [1]::

            eᵢ⁽ˡᵒᵒ⁾ := (φ(X)ᵢβ̂ - yᵢ) / (1 - hᵢ)  where  φ(X)ᵢ is the i'th row of φ(X)
            hᵢ := sᵢφ(X)ᵢ (γC + φ(X)'SSφ(X))⁻¹ sᵢφ(X)ᵢ'
                = sᵢφ(X)ᵢ Q(γ𝕀 + Λ)⁻¹Q⁻¹C⁻¹    sᵢφ(X)ᵢ'

        The entries of hᵢ are also rational polynomials of γ: sᵢφ(X)ᵢQ diag([rᵢ(γ)]ᵢ) Q⁻¹C⁻¹φ(X)ᵢ'.

        We find the γ that optimises the weighted mean absolute leave-one-out error s'|e⁽ˡᵒᵒ⁾| by
        sampling sufficient γs and picking the best one.

        References
        ----------
        [1] https://robjhyndman.com/hyndsight/loocv-linear-models/
        """
        # Normalise the sample weights.
        s = s / np.sum(s)
        # Compute the GEVD φ(X)'SSφ(X)Q = CQΛ so that QΛQ⁻¹ = C⁻¹φ(X)'SSφ(X).
        Sφ = s[:, np.newaxis] * φ
        A = Sφ.conj().T @ Sφ
        A = (A + A.conj().T) / 2  # Ensure A is fully Hermitian.
        c = np.diag(C)
        C_is_diagonal = np.all(np.diag(c) == C)
        C = C / np.mean(np.abs(c)) / φ.size  # Normalise C.
        c = c / np.mean(np.abs(c)) / φ.size  # Normalise c.
        if C_is_diagonal:
            λ, Q = eigh((1 / c[:, np.newaxis]) * A)  # Scipy's eigh is faster for complex matrices.
            CQ_inv = Q.conj().T * (1 / c[np.newaxis, :])
        else:
            λ, Q = eigh(a=A, b=C)
            CQ_lu = lu_factor(C @ Q)
        # Compute the optimal parameters β̂ = Q(γI + Λ)⁻¹Q⁻¹C⁻¹φ(X)'SSy as a function of γ. We can
        # evaluate β̂(γ) as β̂ @ (1 / (γ + λ)) for a given γ.
        φSTSy = Sφ.conj().T @ (s * y)
        β̂ = (
            Q * (CQ_inv @ φSTSy)[np.newaxis, :]
            if C_is_diagonal
            else Q * lu_solve(CQ_lu, φSTSy)[np.newaxis, :]
        )
        # Compute part of the leave-one-out residual numerator sᵢ(yᵢ - φ(X)ᵢβ̂).
        φβ̂ = np.real(φ @ β̂)
        # Compute the leverage part hᵢ of the leave-one-out residual denominator 1 - hᵢ.
        h = (
            np.real(Sφ @ Q * (CQ_inv @ Sφ.conj().T).T)
            if C_is_diagonal
            else np.real(Sφ @ Q * lu_solve(CQ_lu, Sφ.conj().T).T)
        )
        # After np.real, arrays are _neither_ C nor F contiguous, which destroys performance.
        φβ̂ = np.ascontiguousarray(φβ̂)
        h = np.ascontiguousarray(h)
        # Evaluate the unweighted leave-one-out residuals for a set of γs with two matrix-matrix
        # products and pick the best solution.
        self.γs_: FloatVector[F] = np.logspace(np.log10(1e-6), np.log10(20), 1024, dtype=y.dtype)
        rγ = 1 / (self.γs_[np.newaxis, :] + λ[:, np.newaxis])
        with np.errstate(divide="ignore", invalid="ignore"):
            loo_residuals = (φβ̂ @ rγ - y[:, np.newaxis]) / (1 - h @ rγ)
            ŷ_loo = y[:, np.newaxis] + loo_residuals
        # In the case of binary classification, clip overly positive and overly negative
        # predictions' residuals to 0 when the labels are positive and negative, respectively.
        if self._estimator_type == "classifier":
            loo_residuals[(y > 0)[:, np.newaxis] & (loo_residuals > 0)] = 0
            loo_residuals[(y < 0)[:, np.newaxis] & (loo_residuals < 0)] = 0
        # Select γ that minimises the number of LOO misclassifications, the degree to which
        # LOO instances are misclassified, and the weighted absolute LOO error.
        self.loo_errors_γs_ = s @ np.abs(loo_residuals)
        optimum = np.argmin(
            s @ (np.abs(loo_residuals) >= 1)
            + s @ np.maximum(0, np.abs(loo_residuals) - 1)
            + self.loo_errors_γs_
            if self._estimator_type == "classifier"
            else self.loo_errors_γs_
        )
        # Store the leave-one-out residuals, leverage, error, and score.
        self.loo_residuals_ = loo_residuals[:, optimum]
        self.loo_ŷ_ = y + self.loo_residuals_
        self.loo_leverage_ = h @ rγ[:, optimum]
        self.loo_error_ = self.loo_errors_γs_[optimum]
        if self._estimator_type == "classifier":
            self.loo_score_ = accuracy_score(y, np.sign(ŷ_loo[:, optimum]), sample_weight=s)
        elif self._estimator_type == "regressor":
            self.loo_score_ = r2_score(y, ŷ_loo[:, optimum], sample_weight=s)
        β̂, γ = β̂ @ rγ[:, optimum], self.γs_[optimum]
        # Resolve the linear system for better accuracy.
        self.L_ = cho_factor(γ * C + A)
        β̂ = cho_solve(self.L_, φSTSy)
        self.residuals_ = np.real(φ @ β̂) - y
        if self._estimator_type == "classifier":
            self.residuals_[(y > 0) & (self.residuals_ > 0)] = 0
            self.residuals_[(y < 0) & (self.residuals_ < 0)] = 0
        # Compute the leave-one-out nonconformity with the Sherman-Morrison formula.
        σ2 = np.real(np.sum(φ * cho_solve(self.L_, φ.conj().T).T, axis=1))
        σ2 = np.ascontiguousarray(σ2)
        loo_σ2 = σ2 + (s * σ2) ** 2 / (1 - self.loo_leverage_)
        self.loo_nonconformity_ = np.sqrt(loo_σ2)
        # TODO: Print warning if optimal γ is found at the edge.
        return β̂, γ

    def _optimize_α̂_γ(
        self,
        X: FloatMatrix[F],
        y: FloatVector[F],
        s: FloatVector[F],
        ρ: float = 1.0,
    ) -> tuple[FloatVector[F], float]:
        """Find the dual solution to argmin ℒ(e,β̂,b,α̂).

        The Lagrangian is defined as::

            ℒ(e,β̂,b,α̂) := 1/(2γρ) e'S²e + 1/2 (β̂'β̂ + b²) + (1-ρ)/(2ρ) α̂'Cα̂ - α̂'(φ(X)β̂ + b - y - e)

        where γ determines the weight of the regularisation terms β̂'β̂ and α̂'Cα̂, which maximise the
        margin and minimise the complexity of the prediction surface, respectively, and ρ determines
        the trade-off between these two regularisation terms. The residuals e are defined as
        e := φ(X)β̂ + b - y.

        There are two differences w.r.t. the classic LS-SVM formulation: we regularise b in addition
        to β̂, and we add a regularisation term for the complexity of the prediction surface of the
        form α̂'Cα̂. Furthermore, we assume that S = diag(s) and that C is symmetric.

        Setting the gradient of the Lagrangian to zero yields::

            ∂ℒ/∂e = 1/(γρ) S²e + α̂ = 0 => e = -γρ S⁻²α̂
            ∂ℒ/∂β̂ = β̂ - φ(X)'α̂ = 0 => β̂ = φ(X)'α̂
            ∂ℒ/∂b = b - 1'α̂ = 0 => b = 1'α̂
            ∂ℒ/∂α̂ = (1-ρ)/ρ Cα̂ - φ(X)β̂ - b + y + e = 0 => [φ(X)φ(X)' + 11' - (1-ρ)/ρ C + γρS⁻²]α̂ = y

        Let K := φ(X)φ(X)' + 11' - (1-ρ)/ρ C, then we can solve for α̂(γ)::

            (γρS⁻² + K) α̂ = y  and  ŷ(x) := k(x, X)α̂ + b
            S⁻¹(γρ𝕀 + SKS)S⁻¹ α̂ = y
            S⁻¹(γρQQ⁻¹ + QΛQ⁻¹)S⁻¹ α̂ = y  where  SKS Q = QΛ
            α̂ = SQ(γρ𝕀 + Λ)⁻¹Q⁻¹S y

        The entries of α̂ are rational polynomials of γ: Q diag([rᵢ(γ)]ᵢ) Q⁻¹D⁻¹ y.

        Next, we derive the unweighted leave-one-out residuals e⁽ˡᵒᵒ⁾ by analogy to [1]. First, we
        define F := φ(X)φ(X)' + 11', G := -(1-ρ)/ρ C + γρS⁻², and H := (F + G)⁻¹ so that::

            (F + G) α̂ = y
            [f₁₁+g₁₁ f₁'+g₁'; f₁+g₁ F₁+G₁] α̂ = y
            α̂ = Hy = [h₁₁ h₁'; h₁ H₁] y
            h₁₁ := 1/(f₁₁+g₁₁ - (f₁'+g₁')(F₁+G₁)⁻¹(f₁+g₁))
            h₁  := -h₁₁(F₁+G₁)⁻¹(f₁+g₁)
            ŷ₁⁽⁻¹⁾ := f₁' α̂⁽⁻¹⁾
                   = f₁' (F₁+G₁)⁻¹ y⁽⁻¹⁾
                   = f₁' (F₁+G₁)⁻¹ [f₁+g₁ F₁+G₁] α̂
                   = f₁'[-h₁/h₁₁ 𝕀] α̂
            y₁ = [f₁₁+g₁₁ f₁'+g₁'] α̂
            e₁⁽ˡᵒᵒ⁾ := ŷ₁⁽⁻¹⁾ - y₁ = f₁'[-h₁/h₁₁ 𝕀] α̂ - y₁

        We find the γ that optimises the weighted mean absolute leave-one-out error s'|e⁽ˡᵒᵒ⁾| by
        sampling sufficient γs and picking the best one.

        References
        ----------
        [1] http://theoval.cmp.uea.ac.uk/publications/pdf/ijcnn2006a.pdf
        """
        # Normalise the sample weights.
        s = s / np.sum(s)
        sn = s / np.median(np.abs(s))
        # Construct the regularisation matrix C that penalises the complexity of the prediction
        # surface. TODO: Document the derivation of this term.
        gamma = 0.5
        C = np.sqrt(rbf_kernel(X, gamma=gamma)) * (
            1 - euclidean_distances(X, squared=True) * (gamma / X.shape[1])
        )
        # Construct F := φ(X)φ(X)' + 11'.
        F = rbf_kernel(X, gamma=0.5) + np.ones(X.shape[0], dtype=X.dtype)
        # Construct D⁻¹K := 1/ρ S² [φ(X)φ(X)' + 11' - (1-ρ)/ρ C].
        K = F - (1 - ρ) / ρ * C
        # Compute the EVD SKS Q = QΛ.
        λ, Q = np.linalg.eigh(sn[:, np.newaxis] * K * sn[np.newaxis, :])
        # Compute the optimal parameters â = SQ(γρI + Λ)⁻¹Q⁻¹S y as a function of γ. We can evaluate
        # â(γ) as α̂ @ (1 / (γρ + λ)) for a given γ.
        α̂ = (sn[:, np.newaxis] * Q) * (Q.conj().T @ (sn * y))[np.newaxis, :]
        # Evaluate the unweighted leave-one-out residuals for a set of γs and pick the best one.
        self.γs_ = np.logspace(np.log10(1e-6), np.log10(20), 128, dtype=X.dtype)
        # Compute the leave-one-out predictions ŷ₁⁽⁻¹⁾ = f₁'[-h₁/h₁₁ 𝕀] α̂.
        H_loo = np.einsum(  # Compute H := SQ(γρI + Λ)⁻¹Q⁻¹S as a function of γ.
            "ij,gj,jk->igk",
            sn[:, np.newaxis] * Q,
            1 / (self.γs_[:, np.newaxis] * ρ + λ[np.newaxis, :]),
            Q.conj().T * sn[np.newaxis, :],
            optimize="optimal",
        )
        for g in range(H_loo.shape[1]):
            h = np.diag(H_loo[:, g, :]).copy()
            h[h == 0] = np.finfo(X.dtype).eps  # Avoid division by zero.
            H_loo[:, g, :] = H_loo[:, g, :] / -h[:, np.newaxis]
        F_loo = F.copy()
        np.fill_diagonal(F_loo, 0)
        α̂_loo = α̂ @ (1 / (self.γs_[np.newaxis, :] * ρ + λ[:, np.newaxis]))
        ŷ_loo = np.sum(F_loo[:, np.newaxis, :] * H_loo, axis=2) * α̂_loo + F_loo @ α̂_loo
        loo_residuals = ŷ_loo - y[:, np.newaxis]
        # In the case of binary classification, clip overly positive and overly negative
        # predictions' residuals to 0 when the labels are positive and negative, respectively.
        if self._estimator_type == "classifier":
            loo_residuals[(y > 0)[:, np.newaxis] & (loo_residuals > 0)] = 0
            loo_residuals[(y < 0)[:, np.newaxis] & (loo_residuals < 0)] = 0
        # Select γ that minimises the number of LOO misclassifications, the degree to which
        # LOO instances are misclassified, and the weighted absolute LOO error.
        self.loo_errors_γs_ = s @ np.abs(loo_residuals)
        optimum = np.argmin(
            s @ (np.abs(loo_residuals) >= 1)
            + s @ np.maximum(0, np.abs(loo_residuals) - 1)
            + self.loo_errors_γs_
            if self._estimator_type == "classifier"
            else self.loo_errors_γs_
        )
        # Store the leave-one-out residuals, leverage, error, and score.
        self.loo_residuals_ = loo_residuals[:, optimum]
        self.loo_ŷ_ = y + self.loo_residuals_
        self.loo_error_ = self.loo_errors_γs_[optimum]
        if self._estimator_type == "classifier":
            self.loo_score_ = accuracy_score(y, np.sign(ŷ_loo[:, optimum]), sample_weight=s)
        elif self._estimator_type == "regressor":
            self.loo_score_ = r2_score(y, ŷ_loo[:, optimum], sample_weight=s)
        α̂, γ = α̂_loo[:, optimum], self.γs_[optimum]
        # Resolve the linear system for better accuracy.
        self.L_ = cho_factor(γ * ρ * np.diag(sn**-2) + K)
        α̂ = cho_solve(self.L_, y)
        self.residuals_ = F @ α̂ - y
        if self._estimator_type == "classifier":
            self.residuals_[(y > 0) & (self.residuals_ > 0)] = 0
            self.residuals_[(y < 0) & (self.residuals_ < 0)] = 0
        # Compute the nonconformity. TODO: Apply a leave-one-out correction.
        K = rbf_kernel(X, X, gamma=0.5)
        σ2 = 1.0 - np.sum(K * cho_solve(self.L_, K.T).T, axis=1)
        self.loo_nonconformity_ = np.sqrt(σ2)
        # TODO: Print warning if optimal γ is found at the edge.
        return α̂, γ

    def fit(
        self, X: FloatMatrix[F], y: GenericVector, sample_weight: FloatVector[F] | None = None
    ) -> "NeoLSSVM":
        """Fit this predictor."""
        # Remove singleton dimensions from y and validate input.
        X, y = check_X_y(X, y, dtype=(np.float64, np.float32))
        y = np.ravel(np.asarray(y))
        self.n_features_in_ = X.shape[1]
        # Store the target's dtype for prediction.
        self.y_dtype_: npt.DTypeLike = y.dtype
        # Use uniform sample weights if none are provided.
        sample_weight_ = (
            np.ones(y.shape, X.dtype)
            if sample_weight is None
            else np.ravel(np.asarray(sample_weight)).astype(X.dtype)
        )
        check_consistent_length(y, sample_weight_)
        # Learn the type of task from the target. Set the `_estimator_type` and store the `classes_`
        # attribute if this is a classification task [1].
        # [1] https://scikit-learn.org/stable/glossary.html#term-classifiers
        y_: FloatVector[F]
        unique_y = np.unique(y)
        inferred_estimator_type = None
        if len(unique_y) == 2:  # noqa: PLR2004
            inferred_estimator_type = "classifier"
        elif (
            np.issubdtype(y.dtype, np.number)
            or np.issubdtype(y.dtype, np.datetime64)
            or np.issubdtype(y.dtype, np.timedelta64)
        ):
            inferred_estimator_type = "regressor"
        self._estimator_type: str | None = (
            inferred_estimator_type if self.estimator_type == "auto" else self.estimator_type
        )
        if self._estimator_type == "classifier":
            self.classes_: GenericVector = unique_y
            negatives = y == self.classes_[0]
            y_ = np.ones(y.shape, dtype=X.dtype)
            y_[negatives] = -1
        elif self._estimator_type == "regressor":
            y_ = y.astype(X.dtype)
        else:
            message = "Target type not supported"
            raise ValueError(message)
        # Determine whether we want to solve this in the primal or dual space.
        self.dual_ = X.shape[0] <= 1024 if self.dual == "auto" else self.dual  # noqa: PLR2004
        self.primal_ = not self.dual_
        # Learn an optimal distance metric for the primal or dual space and apply it to the feature
        # matrix X.
        if self.primal_:
            self.primal_feature_map_ = clone(
                OrthogonalRandomFourierFeatures()
                if self.primal_feature_map == "auto"
                else self.primal_feature_map
            )
            self.primal_feature_map_.fit(X, y_, sample_weight_)
            φ = self.primal_feature_map_.transform(X)
        else:
            nz_weight = sample_weight_ > 0
            X, y_, sample_weight_ = X[nz_weight], y_[nz_weight], sample_weight_[nz_weight]
            self.dual_feature_map_ = clone(
                AffineSeparator() if self.dual_feature_map == "auto" else self.dual_feature_map
            )
            self.dual_feature_map_.fit(X, y_, sample_weight_)
            self.X_ = self.dual_feature_map_.transform(X)
        # Solve the primal or dual system. We optimise the following sub-objectives for the weights
        # β̂ and hyperparameter γ:
        #   1. Minimal mean squared error on training set: ||S(y - φ(X)β̂)||²
        #   2. Maximal margin: ||β̂||².
        #   3. Minimal complexity of the prediction surface: ∫||∇ₓφ(x)'β̂||²dx
        if self.primal_:
            C = self.primal_feature_map_.complexity_matrix.astype(φ.dtype)
            self.β̂_, self.γ_ = self._optimize_β̂_γ(φ=φ, y=y_, s=sample_weight_, C=C)
        else:
            self.α̂_, self.γ_ = self._optimize_α̂_γ(X=self.X_, y=y_, s=sample_weight_)
        # Calibrate probabilities with isotonic regression on the leave-one-out predictions.
        if self._estimator_type == "classifier":
            self.predict_proba_calibrator_ = IsotonicRegression(
                out_of_bounds="clip", y_min=0, y_max=1, increasing=True
            )
            target = np.zeros_like(y_)
            target[y_ == np.max(y_)] = 1.0
            self.predict_proba_calibrator_.fit(self.loo_ŷ_, target, sample_weight_)
        # Lazily fit conformal predictors as quantile regression models that predict the lower and
        # upper bounds of the (relative) leave-one-out residuals.
        self.conformal_regressors_: dict[str, dict[float, QuantileRegressor]] = {
            "Δ⁺": {},
            "Δ⁻": {},
            "Δ⁺/ŷ": {},
            "Δ⁻/ŷ": {},
        }
        return self

    def nonconformity_measure(self, X: FloatMatrix[F]) -> FloatVector[F]:
        """Compute the nonconformity of a set of examples."""
        # Estimate the nonconformity as the variance of this model's Gaussian Process.
        σ2: FloatVector[F]
        if self.primal_:
            # If β̂ := (LL')⁻¹ y* and cov(y*) := LL', then cov(β̂) = cov((LL')⁻¹ y*) = (LL')⁻¹
            # assuming 𝔼(β̂) = 0. It follows that cov(ŷ(x)) = cov(φ(x)'β̂) = φ(x)'(LL')⁻¹φ(x).
            φH = cast(KernelApproximatingFeatureMap, self.primal_feature_map_).transform(X)
            σ2 = np.real(np.sum(φH * cho_solve(self.L_, φH.conj().T).T, axis=1))
            σ2 = np.ascontiguousarray(σ2)
        else:
            # Compute the cov(ŷ(x)) as K(x, x) − K(x, X) (LL')⁻¹ K(X, x). TODO: Document derivation.
            X = cast(AffineFeatureMap, self.dual_feature_map_).transform(X)
            K = rbf_kernel(X, self.X_, gamma=0.5)
            σ2 = 1.0 - np.sum(K * cho_solve(self.L_, K.T).T, axis=1)
        # Convert the variance to a standard deviation.
        σ = np.sqrt(σ2)
        return σ

    def predict_confidence_interval(
        self, X: FloatMatrix[F], *, confidence_level: float = 0.95
    ) -> FloatMatrix[F] | FloatTensor[F]:
        # Determine the quantiles at the edge of the confidence interval.
        quantile = 1 - (1 - confidence_level) / 2
        # Lazily fit any missing conformal regressors.
        # TODO: Perhaps exclude samples that were used in the feature map.
        # TODO: Perhaps enforce a positive slope for the nonconformity measure.
        for target_type in ("Δ⁺", "Δ⁻", "Δ⁺/ŷ", "Δ⁻/ŷ"):
            quantile_regressors = self.conformal_regressors_[target_type]
            if quantile not in quantile_regressors:
                sgn = (self.loo_residuals_ > 0) if "⁺" in target_type else (self.loo_residuals_ < 0)
                eps = np.finfo(self.loo_ŷ_.dtype).eps
                X_qr = np.hstack(
                    [
                        self.loo_nonconformity_[sgn, np.newaxis],
                        self.loo_ŷ_[sgn, np.newaxis],
                        np.abs(self.loo_ŷ_[sgn, np.newaxis]),
                        np.sign(self.loo_ŷ_[sgn, np.newaxis]),
                    ]
                )
                y_qr = (
                    np.abs(self.loo_residuals_[sgn]) / np.maximum(np.abs(self.loo_ŷ_)[sgn], eps)
                    if "/ŷ" in target_type
                    else np.abs(self.loo_residuals_[sgn])
                )
                quantile_regressors[quantile] = QuantileRegressor(
                    quantile=quantile, alpha=np.sqrt(eps), solver="highs"
                ).fit(X_qr, y_qr)
        # Predict the confidence interval for the nonconformity measure.
        ŷ = self.decision_function(X)
        X_qr = np.hstack(
            [
                self.nonconformity_measure(X)[:, np.newaxis],
                ŷ[:, np.newaxis],
                np.abs(ŷ[:, np.newaxis]),
                np.sign(ŷ[:, np.newaxis]),
            ]
        )
        Δ_lower = np.minimum(
            self.conformal_regressors_["Δ⁻"][quantile].predict(X_qr),
            np.abs(ŷ) * self.conformal_regressors_["Δ⁻/ŷ"][quantile].predict(X_qr),
        )
        Δ_upper = np.minimum(
            self.conformal_regressors_["Δ⁺"][quantile].predict(X_qr),
            np.abs(ŷ) * self.conformal_regressors_["Δ⁺/ŷ"][quantile].predict(X_qr),
        )
        Δ_lower, Δ_upper = np.maximum(0, Δ_lower), np.maximum(0, Δ_upper)
        # Assemble the confidence interval.
        C = np.hstack(((ŷ - Δ_lower)[:, np.newaxis], (ŷ + Δ_upper)[:, np.newaxis]))
        # In case of classification, convert the decision function values to probabilities.
        if self._estimator_type == "classifier":
            C = np.hstack(
                [
                    self.predict_proba_calibrator_.transform(C[:, 0])[:, np.newaxis],
                    self.predict_proba_calibrator_.transform(C[:, 1])[:, np.newaxis],
                ]
            )
            C = np.dstack([1 - C[:, ::-1], C])
        return C

    def decision_function(self, X: FloatMatrix[F]) -> FloatVector[F]:
        """Evaluate this predictor's prediction function."""
        # Compute the point predictions ŷ(X).
        ŷ: FloatVector[F]
        if self.primal_:
            # Apply the feature map φ and predict as ŷ(x) := φ(x)'β̂.
            φ = cast(KernelApproximatingFeatureMap, self.primal_feature_map_).transform(X)
            ŷ = np.real(φ @ self.β̂_)
            ŷ = np.ascontiguousarray(ŷ)
        else:
            # Apply an affine transformation to X, then predict as ŷ(x) := k(x, X) α̂ + 1'α̂.
            X = cast(AffineFeatureMap, self.dual_feature_map_).transform(X)
            K = rbf_kernel(X, self.X_, gamma=0.5)
            b = np.sum(self.α̂_)
            ŷ = K @ self.α̂_ + b
        return ŷ

    def predict(self, X: FloatMatrix[F]) -> GenericVector:
        """Predict the output on a given dataset."""
        # Compute the point predictions ŷ(X).
        ŷ_df = self.decision_function(X)
        if self._estimator_type == "classifier":
            # For binary classification, round to the nearest class label. When the decision
            # function is 0, we assign a negative class label [1].
            # [1] https://scikit-learn.org/stable/glossary.html#term-decision_function
            ŷ_df = np.sign(ŷ_df)
            ŷ_df[ŷ_df == 0] = -1
            # Remap to the original class labels.
            ŷ = self.classes_[((ŷ_df + 1) // 2).astype(np.intp)]
        elif self._estimator_type == "regressor":
            # The decision function is the point prediction.
            ŷ = ŷ_df
        # Map back to the training target dtype.
        ŷ = ŷ.astype(self.y_dtype_)
        return ŷ

    def predict_proba(
        self,
        X: FloatMatrix[F],
        *,
        confidence_interval: bool = False,
        confidence_level: float = 0.95,
    ) -> FloatVector[F] | FloatMatrix[F] | FloatTensor[F]:
        """Predict the class probability or confidence interval."""
        if confidence_interval:
            # Return the confidence interval for classification or regression.
            C = self.predict_confidence_interval(X, confidence_level=confidence_level)
            return C
        if self._estimator_type == "classifier":
            # Return the class probabilities for classification.
            ŷ_classification = self.decision_function(X)
            p = self.predict_proba_calibrator_.transform(ŷ_classification)
            P = np.hstack([1 - p[:, np.newaxis], p[:, np.newaxis]])
        else:
            # Return the point predictions for regression.
            ŷ_regression = self.predict(X)
            P = ŷ_regression
        return P

    def score(
        self, X: FloatMatrix[F], y: GenericVector, sample_weight: FloatVector[F] | None = None
    ) -> float:
        """Compute the accuracy or R² of this classifier or regressor."""
        ŷ = self.predict(X)
        score: float
        if self._estimator_type == "classifier":
            score = accuracy_score(y, ŷ, sample_weight=sample_weight)
        elif self._estimator_type == "regressor":
            # Cast to a numeric dtype in case the target is a datetime or timedelta.
            score = r2_score(
                y.astype(np.float64), ŷ.astype(np.float64), sample_weight=sample_weight
            )
        return score

    def _more_tags(self) -> dict[str, Any]:
        # https://scikit-learn.org/stable/developers/develop.html#estimator-tags
        return {"binary_only": True, "requires_y": True}



"""Affine separator."""

from typing import TypeVar, cast

import numpy as np
from sklearn.utils import check_consistent_length, check_random_state, check_X_y

from neo_ls_svm._affine_feature_map import AffineFeatureMap
from neo_ls_svm._affine_normalizer import AffineNormalizer
from neo_ls_svm._quantizer import sample_bins_quantized_ecdf
from neo_ls_svm._typing import FloatMatrix, FloatVector

F = TypeVar("F", np.float32, np.float64)


def pairwise_distances(X: FloatMatrix[F], Y: FloatMatrix[F]) -> FloatMatrix[F]:
    """Compute the pairwise squared Euclidian distances between the rows of X and Y."""
    d: FloatMatrix[F] = (
        np.sum(X * X, axis=1, keepdims=True) - 2 * X @ Y.T + np.sum(Y * Y, axis=1, keepdims=True).T
    )
    return d


def nearest_neighbours(X: FloatMatrix[F], Y: FloatMatrix[F]) -> FloatMatrix[F]:
    """Find the rows in Y that are nearest to each row in X."""
    pd = pairwise_distances(X, Y)
    idx = np.argmin(pd, axis=1, keepdims=True)
    nn = np.take_along_axis(Y, idx, axis=0)
    return nn


def _faster_svd(X: FloatMatrix[F]) -> tuple[FloatVector[F], FloatMatrix[F]]:
    """Faster algorithm to compute the the right singular vectors of a matrix.

    This function is equivalent to:

        _, s, VH = np.linalg.svd(X, full_matrices=False)
        V = VH.conj().T
    """
    if X.shape[0] >= X.shape[1]:  # Tall and skinny matrix X.
        e, V = np.linalg.eigh(X.conj().T @ X)
        s = np.sqrt(np.abs(e))[::-1]  # eigh returns eigenvalues in ascending order.
        V = V[:, ::-1]
    else:  # Fat and wide matrix X.
        e, U = np.linalg.eigh(X @ X.conj().T)
        s = np.sqrt(np.abs(e))[::-1]
        U = U[:, ::-1]
        nonzero_sv = s > 0
        s, U = s[nonzero_sv], U[:, nonzero_sv]
        V = (X.conj().T @ U) / s[np.newaxis, :]
    return s, V


class AffineSeparator(AffineNormalizer):
    """Affine separator.

    Applies the transformation (x - shift) @ diag(1 / scale) @ A to the input row x so that:

        1. shift centers the features x so that the class bins are optimally separated,
        2. scale scales the shifted features x so that the difference between two samples from
           different class bins is the separability between those class bins, and
        3. A transforms the shifted and scaled features x to optimally separate the labels y.

    The shift and scale are computed by AffineNormalizer.

    To compute the matrix A, the samples are first quantized into class bins. Then, for each class
    bin we select samples P on the edge of the class bin and their (approximate) nearest neighbours
    Q on the edge of all other class bins (the class bin's complement). We want to find a matrix A
    that separates rows p and q of the matrices P and Q, respectively. In other words, we're looking
    for a matrix A that maximizes the distance ||pA - qA||. If UΣV' is the SVD of P-Q, then the
    first columns of V will maximize that distance. This process is repeated for all of the class
    bins, and the final A is a horizontal concatenation of all right singular vectors.

    Next, we'll optimally scale A for Gaussian kernel-based methods with a scalar λ. If f(λA) is the
    expected squared distance between a pair of inter-bin examples, and g(λA) is the expected
    squared distance of a pair of intra-bin examples, then we know that the Gaussian kernel will
    evaluate to exp(-f(λA)/2) and exp(-g(λA)/2) in these two cases, respectively. We then choose the
    scalar λ so that it optimally separates the inter- and intra-bin samples as follows:

        λ := argmin exp(-f(λA)/2) - exp(-g(λA)/2)
        d/dλ exp(-f(λA)/2) - exp(-g(λA)/2) = 0
        d/dλ exp(-λ²f(A)/2) - exp(-λ²g(A)/2) = 0
        f exp(-λ²f/2) = g exp(-λ²g/2)
        exp(λ²(f-g)/2) = f/g
        λ²(f-g)/2 = log(f/g)
        λ = sqrt(2 log(f/g) / (f-g))
    """

    def __init__(  # noqa: PLR0913
        self,
        *,
        append_features: bool = False,
        rank_threshold: float = 2e-2,
        edge_sample_size: int = 384,
        edge_search_multiplier: int = 4,
        random_state: int | np.random.RandomState | None = 42,
    ) -> None:
        self.shift = 0.0
        self.scale = 1.0
        self.A = None
        self.append_features = append_features
        self.rank_threshold = rank_threshold
        self.edge_sample_size = edge_sample_size
        self.edge_search_multiplier = edge_search_multiplier
        self.random_state = random_state

    def fit(
        self,
        X: FloatMatrix[F],
        y: FloatVector[F] | None = None,
        sample_weight: FloatVector[F] | None = None,
    ) -> AffineFeatureMap:
        """Fit this transformer."""
        # Validate X and y.
        assert y is not None
        X, y = check_X_y(X, y, dtype=(np.float64, np.float32))
        y = np.ravel(np.asarray(y)).astype(X.dtype)
        # Learn an optimal shift and scale.
        super().fit(X, y, sample_weight)
        # Transform X with the computed shift and scale.
        X = super().transform(X)  # Note: this won't append features yet because A is still None.
        # Use uniform sample weights if none are provided.
        sample_weight_ = cast(
            FloatVector[F],
            np.ones(y.shape) if sample_weight is None else np.ravel(np.asarray(sample_weight)),
        ).astype(y.dtype)
        check_consistent_length(y, sample_weight_)
        # Separate X into bins according to y.
        y_quantized = sample_bins_quantized_ecdf(y)
        bins = [y_quantized == i for i in range(np.min(y_quantized), np.max(y_quantized) + 1)]
        X_bins = [X[bin, :] for bin in bins]  # noqa: A001
        n_bins = [np.sum(sample_weight_[bin]) for bin in bins]  # noqa: A001
        s_bins = [sample_weight_[np.newaxis, bin] / np.sum(sample_weight_[bin]) for bin in bins]  # noqa: A001
        # Exit early if there is only one bin.
        if len(X_bins) <= 1:
            return self
        # Increase sample size if there are only two bins.
        if len(X_bins) == 2:  # noqa: PLR2004
            self.edge_sample_size = int(self.edge_sample_size * 4 / 3)
        # Learn a tranform A that optimally separates each class bin from the other class bins.
        A_bins, X_bins_edge, X_not_bins_edge = [], [], []
        generator = check_random_state(self.random_state)
        for i in range(len(X_bins)):
            # Generate samples from class bin i.
            idx = generator.choice(
                len(X_bins[i]), size=self.edge_sample_size, p=np.ravel(s_bins[i])
            )
            X_bin_i_sample = X_bins[i][idx, :]
            # Generate samples from the complement of class bin i.
            X_not_bin_i = np.vstack([X_bin for j, X_bin in enumerate(X_bins) if j != i])
            s_not_bin_i = np.hstack([sample_weight_[bin] for j, bin in enumerate(bins) if j != i])  # noqa: A001
            idx = generator.choice(
                len(X_not_bin_i),
                size=self.edge_sample_size * self.edge_search_multiplier,
                p=np.ravel(s_not_bin_i) / np.sum(s_not_bin_i),
            )
            X_not_bin_i_sample = X_not_bin_i[idx, :]
            # Find the samples in the complement of bin i that are closest to bin i.
            X_not_bin_i_edge = nearest_neighbours(X_bin_i_sample, X_not_bin_i_sample)
            X_not_bins_edge.append(X_not_bin_i_edge)
            # Find the samples in bin i that are closest to the nearest neighbours in the bin i's
            # complement.
            idx = generator.choice(
                len(X_bins[i]),
                size=self.edge_sample_size * self.edge_search_multiplier,
                p=np.ravel(s_bins[i]),
            )
            X_bin_i_sample = X_bins[i][idx, :]
            X_bin_i_edge = nearest_neighbours(X_not_bin_i_edge, X_bin_i_sample)
            X_bins_edge.append(X_bin_i_edge)
            # Learn the vectors that maximise the separation between samples from the edge of bin i
            # and the edge of its complement.
            s, V = _faster_svd(X_bin_i_edge - X_not_bin_i_edge)
            rank = np.sum(s > self.rank_threshold * s[0])
            A_bins.append(V[:, :rank])  # TODO: Should these be weighted by the singular values?
        self.A_ = np.hstack(A_bins)
        # Scale A optimally to improve performance for Gaussian kernel-based models.
        inter_bin_distance, intra_bin_distance = 0.0, 0.0
        num_inter_pairs = self.edge_sample_size * (self.edge_sample_size + 1) / 2
        num_intra_pairs = self.edge_sample_size * (self.edge_sample_size - 1) / 2
        for X_bin_edge, X_not_bin_edge, n_bin in zip(
            X_bins_edge, X_not_bins_edge, n_bins, strict=True
        ):
            inter_bin_distance += (
                n_bin
                * np.sum(
                    np.tril(pairwise_distances(X_bin_edge @ self.A_, X_not_bin_edge @ self.A_), k=0)
                )
                / num_inter_pairs
            )
            intra_bin_distance += (
                n_bin
                * np.sum(
                    np.tril(pairwise_distances(X_bin_edge @ self.A_, X_bin_edge @ self.A_), k=-1)
                )
                / num_intra_pairs
            )
        inter_bin_distance /= sum(n_bins)
        intra_bin_distance /= sum(n_bins)
        λ = (
            np.sqrt(
                2
                * np.log(inter_bin_distance / intra_bin_distance)
                / (inter_bin_distance - intra_bin_distance)
            )
            if intra_bin_distance > 0
            else 1
        )
        self.A_ *= λ  # TODO: SVM performance is quite sensitive to this scalar. Can it be improved?
        return self
